///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.8+
// Copyright (C) 2008-2011 fccHandler
// Copyright (C) 1998-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#include "AC3Info.h"


static const unsigned short rate[19] =
{
	 32,  40,  48,  56,  64,  80,  96, 112, 128,
	160, 192, 224, 256, 320, 384, 448, 512, 576, 640
};


// The position of the LFE bit depends on what follows acmod:
// if ((acmod & 1) && (acmod != 1)) cmixlev (2 bits) follows acmod
// if (acmod & 4) surmixlev (2 bits) follows acmod (and cmixlev)

// Bytes 6-7, x = LFE:
//	000x.... ........	1+1, LFE
//  001x.... ........	1/0, LFE
//  010x.... ........	2/0, LFE
//  011..x.. ........	3/0, cmixlev, LFE
//  100..x.. ........	2/1, surmixlev, LFE
//  101....x ........	3/1, cmixlev, surmixlev, LFE
//	110..x.. ........	2/2, surmixlev, LFE
//	111....x ........	3/2, cmixlev, surmixlev, LFE

static const unsigned short lfeon[8] =
{
	0x1000, 0x1000, 0x1000, 0x0400,	0x0400, 0x0100, 0x0400, 0x0100
};


static const unsigned char chan[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };

static const unsigned char halfrate[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3 };


// bytes 4-7
// xx...... ........ ........ ........  fscod (00=48, 01=44.1, 10=32, 11=reserved)
// ..xxxxxx ........ ........ ........  frmsizecod (0 to 37)
// ........ xxxxx... ........ ........  bsid (01000 = version 8)
// ........ .....xxx ........ ........  bsmod
// ........ ........ xxx..... ........  acmod (#channels)
// ........ ........ ...x.x.x ........  lfe (position depends on acmod)

int ac3_sync_info(const unsigned long hdr, int *srate, int *brate, int *chans)
{
	if (hdr < 0xC0000000)
	{
		const int frmsizecod = (hdr >> 24) & 63;

		if (frmsizecod < 38)
		{
			const int bsid = (hdr >> 19) & 31;

			if (bsid < 12)
			{
				const int acmod   = (hdr >> 13) & 7;
				const int half    = halfrate[bsid];
				const int bitrate = rate[frmsizecod >> 1];

				if (chans) *chans = chan[acmod] + !!(hdr & lfeon[acmod]);
				if (brate) *brate = bitrate >> half;

				switch (hdr >> 30)
				{
					case 0:
						if (srate) *srate = 48000 >> half;
						return (4 * bitrate);

					case 1:
						if (srate) *srate = 44100 >> half;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));

					case 2:
						if (srate) *srate = 32000 >> half;
						return (6 * bitrate);
				}
			}
		}
	}
	return 0;
}

