///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.8+
// Copyright (C) 2008-2011 fccHandler
// Copyright (C) 1998-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <malloc.h>

#include "DataVector.h"


DataVector::DataVector(unsigned int item_size)
	: mItemSize(item_size)
{
	mFirst = NULL;
	mLast  = NULL;
	mCount = 0;
}


DataVector::~DataVector()
{
	DataVectorBlock* i;
	DataVectorBlock* j;

	i = mLast;
	while (j = i)
	{
		i = j->Prev;
		delete j;
	}
}


void DataVector::Add(const void *pp)
{
	unsigned int len = mItemSize;

	if (mLast == NULL)
	{
		DataVectorBlock *newb = new DataVectorBlock;

		memcpy(newb->Heap, pp, len);
		newb->Next = NULL;
		newb->Size = len;

		if (mFirst == NULL)
		{
			// New "first" block
			mFirst       = newb;
			mFirst->Prev = NULL;
		}
		else
		{
			// New "last" block
			mLast        = newb;
			mFirst->Next = mLast;
			mLast->Prev  = mFirst;
		}
	}
	else
	{
		const unsigned int remains = (unsigned int)
			(DataVectorBlock::BLOCK_SIZE - mLast->Size);

		if (len > remains)
		{
			// New "last" block
			DataVectorBlock *newb = new DataVectorBlock;

			mLast->Next = newb;
			newb->Prev  = mLast;
			newb->Next  = NULL;
			newb->Size  = 0;

			if (remains > 0)
			{
				memcpy(mLast->Heap + mLast->Size, pp, remains);
				pp = (char *)pp + remains;
				mLast->Size += remains;
				len -= remains;
			}

			mLast = newb;
		}

		memcpy(mLast->Heap + mLast->Size, pp, len);
		mLast->Size += len;
	}

	++mCount;
}


void *DataVector::MakeArray() const
{
	void *arr = malloc(mCount * mItemSize);
	char *ptr = (char *)arr;

	const DataVectorBlock *dvb = mFirst;

	while (dvb != NULL)
	{
		memcpy(ptr, dvb->Heap, dvb->Size);
		ptr += dvb->Size;

		dvb  = dvb->Next;
	}

	return arr;
}

