///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.8+
// Copyright (C) 2008-2011 fccHandler
// Copyright (C) 1998-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"
#include "Resource.h"

#include "WinFile.h"
#include "AC3Info.h"
#include "DataVector.h"
#include "InputFileAC3.h"
#include "AudioSourceAC3.h"


static HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileAC3
//
///////////////////////////////////////////////////////////////////////

InputFileAC3::InputFileAC3(const VDXInputDriverContext& context)
	: mContext(context)
{
	framelist    = NULL;
	fAbort       = false;
	fAc3Wav      = false;
	fIntel       = false;
	frames       = 0;
	mFirstHeader = 0;
}


InputFileAC3::~InputFileAC3()
{
	free(framelist);
}


INT_PTR CALLBACK InputFileAC3::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	InputFileAC3 *thisPtr = (InputFileAC3 *)GetWindowLongPtrA(hDlg, DWLP_USER);

	switch(uMsg) {

		case WM_INITDIALOG:
		{
			SetWindowLongPtrA(hDlg, DWLP_USER, lParam);
			thisPtr = (InputFileAC3 *)lParam;

			SendMessageA(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
			SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
			ShowWindow(hDlg, SW_SHOW);

			return TRUE;
		}

		case WM_TIMER:
		{
			char buf[32];
			__int64 file_cpos = thisPtr->mFile.Pos();
			__int64 file_size = thisPtr->mFile.Size();

			SendMessageA(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETPOS,
				(WPARAM)((file_cpos * 16384) / file_size), 0);

			_snprintf(buf, sizeof(buf), "%luK of %luK",
				(unsigned long)(file_cpos / 1024),
				(unsigned long)((file_size + 512) / 1024)
			);

			SendDlgItemMessageA(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		}
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;

		return TRUE;
	}

	return FALSE;
}


bool InputFileAC3::guiDlgMessageLoop(HWND hDlg)
{
	MSG msg;

	while (PeekMessageA(&msg, NULL, 0, 0, PM_REMOVE) != FALSE)
	{
		if (msg.message == WM_QUIT) return false;

		if (hDlg == NULL || IsWindow(hDlg) == FALSE
			|| IsDialogMessageA(hDlg, &msg) == FALSE)
		{
			TranslateMessage(&msg);
			DispatchMessageA(&msg);
		}
	}

	return true;
}


BOOL CALLBACK InputFileAC3::FindContextWindow(HWND hwnd, LPARAM lParam)
{
	if (GetParent(hwnd) == NULL)
	{
		*(HWND *)lParam = hwnd;
		return FALSE;
	}
	return TRUE;
}


bool InputFileAC3::DetectWav(unsigned long& header, int& frmsize)
{
	// Detect AC3-WAV format

	static const unsigned long SPDIF_SYNC = 0x4E1FF872;

	unsigned long buf[4];
	PCMWAVEFORMAT pcm;

	if (mFile.Read(buf, 12) != 12) goto Abort;

	if (buf[0] != 'FFIR') goto Abort;
	if (buf[2] != 'EVAW') goto Abort;
	
	while (mFile.Read(buf, 8) == 8)
	{
		switch (buf[0])
		{
			case ' tmf':
				if (buf[1] != sizeof(pcm)) goto Abort;
				mFile.Read(&pcm, sizeof(pcm));
				buf[1] = 0;

				if (pcm.wf.wFormatTag != WAVE_FORMAT_PCM) goto Abort;
				if (pcm.wf.nChannels != 2) goto Abort;
				if (pcm.wf.nSamplesPerSec != 44100) goto Abort;
				if (pcm.wf.nAvgBytesPerSec != 176400) goto Abort;
				if (pcm.wf.nBlockAlign != 4) goto Abort;
				if (pcm.wBitsPerSample != 16) goto Abort;
				break;

			case 'atad':
				if (buf[1] < 0x1800) goto Abort;

				mFile.Read(buf, 16);
				if (buf[0] != SPDIF_SYNC) goto Abort;
				if ((buf[1] & 0xFFFF) != 0x0001) goto Abort;
				if ((buf[2] & 0xFFFF) != 0x0B77) goto Abort;

				// Swap to Motorola order
				header = ((buf[3] >> 16) & 0x0000FFFF)
				       | ((buf[3] << 16) & 0xFFFF0000);

				frmsize = ac3_sync_info(header, NULL, NULL, NULL);

				// ........ ........ ........ 00000001  type (1)
				// ........ ........ 00000000 ........  bsmod (0)
				// ........ xxxxxxxx ........ ........  databits LSB
				// xxxxxxxx ........ ........ ........  databits MSB

				// S/PDIF databits, swap to Intel order
				buf[1] = ((buf[1] >> 16) & 0x0000FF00) | (buf[1] & 0x00FF0000);

				// Mystery: why must I divide by 16?
				if (frmsize != (int)(buf[1] / 16)) goto Abort;
				return true;
		}
		
		if (!mFile.Seek(mFile.Pos() + buf[1])) break;
	}

Abort:
	mFile.Seek(0);
	return false;
}


int InputFileAC3::FirstSync(int &srate, int &chans)
{
	unsigned short syncword;
	int            frmsize;
	int            bytes;
	int            limit;

	mFirstHeader = 0;
	frmsize      = 0;

	fAc3Wav  = DetectWav(mFirstHeader, frmsize);
	if (fAc3Wav)
	{
		fIntel = true;
		return frmsize;
	}

	syncword = 0;
	bytes    = 0;

	{
		__int64 remains = mFile.Size() - mFile.Pos();
		limit = (remains > AC3_HUNT_SIZE)? AC3_HUNT_SIZE: (int)remains;
	}

	// Locate the first syncword
	// The smallest possible AC-3 frame is 128 bytes
	
	while (limit >= 128) {

		if (syncword == 0x0B77 || syncword == 0x770B)
		{
			// Got syncword, now get the 6 bytes which follow it

			if (bytes == 6)
			{
				if (syncword == 0x770B)
				{
					// Intel syncword, swap header bytes
					mFirstHeader = ((mFirstHeader >> 8) & 0x00FF00FF)
					             | ((mFirstHeader << 8) & 0xFF00FF00);
				}

				frmsize = ac3_sync_info(mFirstHeader, &srate, NULL, &chans);

				if (frmsize >= 8)
				{
					// Fail if frame is incomplete
					if ((mFile.Pos() + (frmsize - 8)) > mFile.Size()) break;

					// Hooray!
					fIntel = (syncword == 0x770B);
					return frmsize;
				}

				// Go back to hunting for syncword
				syncword = 0;
				bytes    = 0;
			}
			else
			{
				// Accumulate header bytes
				mFirstHeader = (mFirstHeader << 8) + mFile.ReadByte();
				++bytes;
			}
		}
		else
		{
			// Hunt 4K for syncword
			syncword = (syncword << 8) + mFile.ReadByte();
			--limit;
		}
	}

	// No sync found
	return 0;
}


int InputFileAC3::NextSync(int srate, int chans)
{
	unsigned long  header;
	unsigned short syncword;
	unsigned short target;
	int            frmsize;
	int            bytes;
	int            limit;

	header   = 0;
	syncword = 0;
	bytes    = 0;

	{
		__int64 remains = mFile.Size() - mFile.Pos();
		limit = (remains > AC3_HUNT_SIZE)? AC3_HUNT_SIZE: (int)remains;
	}

	// Target syncword
	target = (fIntel)? 0x770B: 0x0B77;

	// Locate the next syncword
	// The smallest possible AC-3 frame is 128 bytes
	
	while (limit >= 128) {

		if (syncword == target)
		{
			// Got syncword, now get the 6 bytes which follow it

			if (bytes == 6)
			{
				int srate2;
				int chans2;

				if (syncword == 0x770B)
				{
					// Intel syncword, swap header bytes
					header = ((header >> 8) & 0x00FF00FF)
						   | ((header << 8) & 0xFF00FF00);
				}

				frmsize = ac3_sync_info(header, &srate2, NULL, &chans2);

				if (frmsize > 0 && srate2 == srate && chans2 == chans)
				{
					// Fail if frame is incomplete
					if ((mFile.Pos() + (frmsize - 8)) > mFile.Size()) break;

					return frmsize;
				}

				// Go back to hunting for syncword
				syncword = 0;
				bytes    = 0;
			}
			else
			{
				// Accumulate header bytes
				header = (header << 8) + mFile.ReadByte();
				++bytes;
			}
		}
		else
		{
			// Hunt 4K for syncword
			syncword = (syncword << 8) + mFile.ReadByte();
			--limit;
		}
	}

	// No sync found
	return 0;
}


void VDXAPIENTRY InputFileAC3::Init(const wchar_t *szFile, IVDXInputOptions *opts)
{
	DataVector dv(sizeof(AC3SampleInfo));

	unsigned long header;
	int frmsize, srate, chans;
	AC3SampleInfo asi;

	HWND hwndProgress;
	HWND hwndParent;

	if (!mFile.Open(szFile))
	{
		mContext.mpCallbacks->SetError("Can't open %ls", szFile);
		return;
	}

	mFirstHeader = 0;
	header   = 0;
	fIntel   = false;
	fAc3Wav  = false;
	fAbort   = false;

	// Locate the first syncword

	frmsize = FirstSync(srate, chans);
	if (frmsize <= 0)
	{
		mContext.mpCallbacks->SetError("Invalid AC-3 stream.");
		return;
	}

	// Get a handle to the main VirtualDub window

	hwndParent = NULL;
	EnumThreadWindows(GetCurrentThreadId(), FindContextWindow, (LPARAM)&hwndParent);
	if (hwndParent == NULL) hwndParent = GetActiveWindow();
	if (hwndParent != NULL) EnableWindow(hwndParent, FALSE);

	free(framelist);
	framelist = NULL;
	frames    = 1;

	asi.size = frmsize;
	asi.pos  = mFile.Pos() - 8;
	dv.Add(&asi);

	mFile.Seek(mFile.Pos() + (fAc3Wav? 0x1800 - 8: frmsize - 8));

	// Pop up parse dialog
	hwndProgress = CreateDialogParamA(
		g_hModule,
		MAKEINTRESOURCEA(IDD_PROGRESS),
		GetActiveWindow(),
		ParseDialogProc,
		(LPARAM)this
	);

	// Add frames
	while (true) {

		// Update the progress dialog
		guiDlgMessageLoop(hwndProgress);
		if (fAbort) break;

		frmsize = NextSync(srate, chans);
		if (frmsize <= 0) break;

		asi.size = frmsize;
		asi.pos  = mFile.Pos() - 8;
		dv.Add(&asi);

		mFile.Seek(mFile.Pos() + (fAc3Wav? 0x1800 - 8: frmsize - 8));
		++frames;
	}

	if (hwndParent   != NULL) EnableWindow(hwndParent, TRUE);
	if (hwndProgress != NULL) DestroyWindow(hwndProgress);

	framelist = (AC3SampleInfo *)(dv.MakeArray());
		
	if (framelist == NULL)
	{
		frames = 0;
		mContext.mpCallbacks->SetErrorOutOfMemory();
	}
}


bool VDXAPIENTRY InputFileAC3::Append(const wchar_t *szFile)
{
	return false;
}


bool VDXAPIENTRY InputFileAC3::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions)
{
	*ppOptions = NULL;
	return false;
}


bool VDXAPIENTRY InputFileAC3::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions)
{
	*ppOptions = NULL;
	return false;
}


void VDXAPIENTRY InputFileAC3::DisplayInfo(VDXHWND hwndParent)
{
}


bool VDXAPIENTRY InputFileAC3::GetVideoSource(int index, IVDXVideoSource **ppVS)
{
	*ppVS = NULL;
	return false;
}


bool VDXAPIENTRY InputFileAC3::GetAudioSource(int index, IVDXAudioSource **ppAS)
{
	*ppAS = NULL;
	if (index != 0) return false;

	*ppAS = new AudioSourceAC3(this);
	if (*ppAS == NULL) return false;

	(*ppAS)->AddRef();
	return true;
}


void InputFileAC3::ByteSwap(void *buffer, int len)
{
	unsigned char *src = (unsigned char *)buffer;
	while (len > 1)
	{
		unsigned char ch = src[0];
		src[0] = src[1];
		src[1] = ch;
		src += 2;
		len -= 2;
	}
}


int InputFileAC3::ReadSample(void *buffer, unsigned long sample)
{
	int actual = 0;

	if (sample < frames)
	{
		__int64 offset = framelist[sample].pos;
		int size = (int)framelist[sample].size;

		if ((offset + size) > mFile.Size())
		{
			mContext.mpCallbacks->SetError("Attempt to read past end of audio");
		}
		else
		{
			mFile.Seek(offset);
			actual = mFile.Read(buffer, size);
			if (fIntel) ByteSwap(buffer, actual);
		}
	}

	return actual;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverAC3
//
///////////////////////////////////////////////////////////////////////

class InputFileDriverAC3 : public vdxunknown<IVDXInputFileDriver>
{
public:
	InputFileDriverAC3(const VDXInputDriverContext& context);

// IVDXInputFileDriver methods
	int     VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool    VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	int     mRefCount;

	const   VDXInputDriverContext& mContext;
};


InputFileDriverAC3::InputFileDriverAC3(const VDXInputDriverContext& context)
	: mRefCount(0)
	, mContext(context)
{
}


int VDXAPIENTRY InputFileDriverAC3::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize)
{
	return -1;
}


bool VDXAPIENTRY InputFileDriverAC3::CreateInputFile(uint32 flags, IVDXInputFile **ppFile)
{
	*ppFile = new InputFileAC3(mContext);
	if (*ppFile == NULL) return false;

	(*ppFile)->AddRef();
	return true;
}


bool VDXAPIENTRY ac3_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver)
{
	*ppDriver = new InputFileDriverAC3(*pContext);
	if (*ppDriver == NULL) return false;

	(*ppDriver)->AddRef();
	return true;
}


const uint8 ac3_sig[] =
{
	0x0B, 0xFF, 0x77, 0xFF,		// syncword
	0, 0, 0, 0,					// crc1
	0, 0,						// fscod, frmsizecod
	0x00, 0xF0					// bsid <= 8
};


// Note: For VirtualDub-1.10.1-test19 and later,
// set kFlagNoOptions (4) to indicate that we do
// not support extended open options.

const VDXInputDriverDefinition ac3_input =
{
	sizeof(VDXInputDriverDefinition),
	VDXInputDriverDefinition::kFlagSupportsAudio | 4,
	0,
	sizeof ac3_sig,
	ac3_sig,
	L"*.ac3",
	L"Dolby Digital (*.ac3)|*.ac3",
	L"ac3",
	ac3_create
};


const VDXPluginInfo ac3_info =
{
	sizeof(VDXPluginInfo),              // size
	L"AC3 plugin",                      // name
	L"fccHandler",                      // author
	L"Loads Dolby Digital audio.",      // description
	0x01090000,                         // (major<<24) + (minor<<16) + build
	kVDXPluginType_Input,               // type
	0,                                  // flags
	10,                                 // version required
	kVDXPlugin_APIVersion,              // version used
	2,                                  // version required
	kVDXPlugin_InputDriverAPIVersion,   // version used
	&ac3_input
};


///////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] =
{
	&ac3_info,
	NULL
};


extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo()
{
	return kPlugins;
}

